#include "map.h"
#include <stdio.h>
#include <stdlib.h>     /* srand, rand */
#include <assert.h>
#include <allegro5/allegro_primitives.h>
#include "util.h"


#include <iostream>


bool FSort(const BasicTile * a_pC1, const BasicTile * a_pC2);


//--------------------------------------------------------------
// Constructor  : BasicTile.BasicTile
// Description  : Constructs a tile
//-------------------------------------------------------------
BasicTile::BasicTile()
{
  m_nOffsetX        =  0;
  m_nOffsetY        =  0;
  m_nBitmapOffsetX  =  -1;
  m_nBitmapOffsetY  =  -1;
  m_pBitmap         =  NULL;
  m_blOccupied      = false;
  m_pParent         = NULL;
  m_dF              = 0;
  m_nDijkstra       = -1;
  m_pBitmap         = NULL;
  m_pEngine         = NULL;
  m_nX              = 0;
  m_nY              = 0;
  m_nTraverseCost   = 1;
  m_blIsPath        = false;


}


//--------------------------------------------------------------
// Destructor   : BasicTile.~BasicTile
// Description  : Destroys a tile
//-------------------------------------------------------------
BasicTile::~BasicTile()
{


}

//--------------------------------------------------------------
// Function     : BasicTile.SetOffset
// Description  : Set's the grabbing offset for a tile
// Virtual      : No
//-------------------------------------------------------------
void BasicTile::SetOffset(int a_nOffsetX,int a_nOffsetY)
{
   m_nBitmapOffsetX = a_nOffsetX;
   m_nBitmapOffsetY = a_nOffsetY;
}


//--------------------------------------------------------------
// Function     : BasicTile.SetBitmap
// Description  : Set's the bitmap for a tile
// Virtual      : No
//-------------------------------------------------------------
void BasicTile::SetBitmap(ALLEGRO_BITMAP  * a_pBitmap)
{
   m_pBitmap = a_pBitmap;
}


//--------------------------------------------------------------
// Function     : BasicTile.Draw
// Description  : Draws a tile at position a_nScreenX, a_nScreenY
// Virtual      : Yes
//-------------------------------------------------------------
void BasicTile::Draw(int a_nScreenX,int a_nScreenY)
{
  a_nScreenX+=m_nOffsetX;
  a_nScreenY+=m_nOffsetY;

  if (m_pBitmap == NULL)
  {
    // No bitmap, draw a placeholder
    ALLEGRO_COLOR  Fill;

    if (m_blOccupied)
    {
        Fill   = al_map_rgb(255,0,0);
    }
    else
    {
        Fill   = al_map_rgb(0,0,0);
    }

    ALLEGRO_COLOR  Border = al_map_rgb(0,255,0);

    al_draw_filled_rectangle(a_nScreenX,a_nScreenY, a_nScreenX+m_pEngine->m_nTilewidth, a_nScreenY+m_pEngine->m_nTileHeight,Fill);
    al_draw_rectangle(a_nScreenX,a_nScreenY, a_nScreenX+m_pEngine->m_nTilewidth, a_nScreenY+m_pEngine->m_nTileHeight,Border,1);

    if (m_blIsPath)
    {
        al_draw_filled_circle(a_nScreenX+(m_pEngine->m_nTilewidth/2),a_nScreenY+(m_pEngine->m_nTileHeight/2),10,Border);
    }
  }
  else
  {
    if ((m_nBitmapOffsetX == -1)&&(m_nBitmapOffsetY == -1))
    {
      // single tile
      al_draw_bitmap (m_pBitmap, a_nScreenX, a_nScreenY,0);
    }
    else
    {
      // tilesheet
      al_draw_bitmap_region (m_pBitmap,m_nBitmapOffsetX, m_nBitmapOffsetY, m_pEngine->m_nTilewidth, m_pEngine->m_nTileHeight, a_nScreenX, a_nScreenY,0);
    }

  }
}


int BasicTile::GetX()
{
    return m_nX;
}

int BasicTile::GetY()
{
    return m_nY;
}



//--------------------------------------------------------------
// Constructor  : BasicEngine::BasicEngine
// Description  : Constructs a tileengine
//-------------------------------------------------------------
BasicEngine::BasicEngine()
{
  m_nTilewidth=64;
  m_nTileHeight=64;
  m_nYStep = m_nTileHeight;
  SCREENCENTER_X = 800;
  SCREENCENTER_Y = 600;
}


//--------------------------------------------------------------
// Destructor   : BasicEngine::~BasicEngine
// Description  : frees all used memory
//-------------------------------------------------------------
BasicEngine::~BasicEngine()
{
  std::vector<std::vector< BasicTile * > >::iterator p=m_lstCell.begin();
  while (p!=m_lstCell.end())
  {
    std::vector< BasicTile * >::iterator q = (*p).begin();
    while (q!=(*p).end())
    {
      BasicTile * hulp = *q;
      (*p).erase(q++);
      delete hulp;
    }
    m_lstCell.erase(p++);
  }
}


//--------------------------------------------------------------
// Function     : BasicEngine.Draw
// Description  : Draws a tilemap centered around a_nCamx,a_nCamy
// Virtual      : Yes
//-------------------------------------------------------------
void BasicEngine::Draw(int a_nCamx,int a_nCamy)
{
  int DrawX;
  int DrawY;
  unsigned int startx,endx,starty,endy;


  //Calculate start and end for nested loops
  if ((a_nCamx-SCREENCENTER_X)<=0)
  {
      startx = 0;
  }
  else
  {
     startx=(a_nCamx-SCREENCENTER_X)/m_nTilewidth;
     if (startx>m_lstCell.size()) startx=m_lstCell.size();
  }

  endx=((a_nCamx+SCREENCENTER_X)/m_nTilewidth)+1;
  if (endx>m_lstCell.size()) endx=m_lstCell.size();


  if ((a_nCamy-SCREENCENTER_Y)<=0)
  {
      starty = 0;
  }
  else
  {
     starty=((a_nCamy-SCREENCENTER_Y)/m_nYStep);
     if (starty>m_lstCell.size()) starty=m_lstCell.size();
  }

  endy=((a_nCamy+SCREENCENTER_Y)/m_nYStep)+1;
  if (endy>m_lstCell.size()) endy=m_lstCell.size();

  int sy=(starty*m_nYStep);
  int h_sx=startx*m_nTilewidth;
  int sx;

  for (unsigned int y=starty;y<endy;y++)
    {
    sx=h_sx;
    for (unsigned int x=startx;x<endx;x++)
        {
          DrawX=int(SCREENCENTER_X+(sx-a_nCamx));
          DrawY=int(SCREENCENTER_Y+(sy-a_nCamy));
          sx+=64;
          if ((x>=0) && (x<(unsigned int)m_nMapWidth) && (y >=0) & (y < (unsigned int)m_nMapHeight))
          {
              m_lstCell[x][y]->Draw(DrawX,DrawY);
          }
          else
          {


          }
        }
    sy+=m_nYStep;
    }

}


//-----------------------------------------------------------------------------
// Function     : BasicEngine.GenerateTest
// Description  : Generates a testmap with Width a_nWidth and height a_nHeight
// Virtual      : Yes
//-----------------------------------------------------------------------------
void BasicEngine::GenerateTest(int a_nWidth, int a_nHeight, ALLEGRO_BITMAP* a_pBitmap)
{
  m_nMapWidth = a_nWidth;
  m_nMapHeight = a_nHeight;

  for (int y=0; y<m_nMapHeight;y++)
  {
    std::vector<BasicTile * > tempvector;
    for (int x=0;x< m_nMapWidth;x++)
    {
      BasicTile * pTile = new BasicTile();
      pTile->m_nX = x;
      pTile->m_nY = y;
      if (a_pBitmap!=NULL)
      {
        pTile->SetBitmap(a_pBitmap);
        if((x==0)||(y==0))
        {
          pTile->SetOffset(0,320);

        }
        else
        {
          pTile->SetOffset(0,256);
        }
      }
      pTile->m_pEngine = this;
      tempvector.push_back(pTile);
    }
    m_lstCell.push_back(tempvector);
  }

  for (int i=0; i< (a_nWidth + a_nHeight*2); i++)
  {
       int x = rand() % a_nWidth;
       int y = rand() % a_nHeight;
       m_lstCell[x][y]->m_blOccupied = true;
  }
}



std::vector<BasicTile *> BasicEngine::FindPath(int a_nStartX, int a_nStartY,int a_nEndX, int a_nEndY)
{
   ClearAstar();
   bool blPathFound = false;
   std::vector<BasicTile *> ResultPath;
   ClearAstar();                                                     // clear astar values
   BasicTile * pEndPath = NULL;                                      // pointer to end of path
   BasicTile * pCurrent = m_lstCell[a_nStartX][a_nStartY];           // set current pointer to start of path
   BasicTile * pStartPath = pCurrent;                                // pointer to start of path
   pCurrent->m_dF = 0;                                               // F cost of start tile is 0
   m_lstOpenList.push_back(pCurrent);                                // push to open list

   while ( (!m_lstOpenList.empty()) && (!blPathFound))               // while open list not empty and no path is found
   {
       pCurrent = m_lstOpenList.front();                             // get the first/next node
       m_lstOpenList.pop_front();                                    // remove first node from list
       m_lstClosedList.push_back(pCurrent);                          // push note to closed list

       std::vector<BasicTile *> vSuccessors;                         // get the possible successors
       vSuccessors = GetAdjectend(pCurrent->m_nX, pCurrent->m_nY);

       for (size_t i=0; i < vSuccessors.size(); i++)                 // loop through each successor
       {
            BasicTile * pSuc = vSuccessors[i];
            if ((pSuc->m_nX == a_nEndX) && (pSuc->m_nY == a_nEndY) && (!blPathFound))  // path found, terminate !
            {
                pSuc->m_pParent = pCurrent;
                pEndPath = pSuc;
                blPathFound = true;
            }
            else if (! IsInClosed(pSuc))                             // ignore if in closed
            {
                if (IsInOpen(pSuc))                                  // if in open list, see what path is shortest
                {
                    double H = abs(a_nEndX - pSuc->m_nX) + abs(a_nEndY - pSuc->m_nY);
                    double dCurF;
                    if (IsDiagonal(pCurrent, pSuc))
                    {
                        dCurF = pCurrent->m_dF + (((double)pSuc->m_nTraverseCost) * 1.8) + H;
                    }
                    else
                    {
                        dCurF = pCurrent->m_dF + pSuc->m_nTraverseCost + H;
                    }

                    if (dCurF < pSuc->m_dF)                            // shorter path, update F and Parent
                    {
                        pSuc->m_dF = dCurF;
                        pSuc->m_pParent = pCurrent;                    // set successor parent to current node
                        m_lstOpenList.sort(FSort);
                    }

                }
                else  // if not in open list, calculate F and add it to the open list
                {
                    double H = abs(a_nEndX - pSuc->m_nX) + abs(a_nEndY - pSuc->m_nY);
                    if (IsDiagonal(pCurrent, pSuc))
                    {
                        pSuc->m_dF = pCurrent->m_dF + (((double)pSuc->m_nTraverseCost) * 1.8) + H;
                    }
                    else
                    {
                         pSuc->m_dF = pCurrent->m_dF + pSuc->m_nTraverseCost + H;
                    }

                    pSuc->m_pParent = pCurrent;      // set successor parent to current node
                    m_lstOpenList.push_back(pSuc);
                    m_lstOpenList.sort(FSort);
                }
            }
        }
   }

   if (blPathFound)  // build the path
   {
       BasicTile * pTile = pEndPath;
       while ((pTile != pStartPath)&&(pTile != NULL))
       {
           pTile->m_blIsPath = true;
           ResultPath.push_back(pTile);
           pTile = pTile->m_pParent;
       }
       pTile->m_blIsPath = true;
       ResultPath.push_back(pTile);
   }

   return ResultPath;
}



void BasicEngine::GenerateDijkstra(int a_nGoalX, int a_nGoalY)
{
    ClearDijkstra();
    std::list<BasicTile *> ToBeVisited;

    BasicTile * pCurrent = m_lstCell[a_nGoalX][a_nGoalY];
    ToBeVisited.push_back(pCurrent);
    pCurrent->m_nDijkstra = 0;

    while (!ToBeVisited.empty())
    {
       pCurrent = ToBeVisited.front();
       ToBeVisited.pop_front();

       std::vector<BasicTile *> vSuccessors;
       vSuccessors = GetAdjectend(pCurrent->m_nX, pCurrent->m_nY);
       for (size_t i=0; i < vSuccessors.size(); i++)
       {
           if (vSuccessors[i]->m_nDijkstra < 0)
           {
               vSuccessors[i]->m_nDijkstra = pCurrent->m_nDijkstra + vSuccessors[i]->m_nTraverseCost;
               BasicTile * pSuc = vSuccessors[i];
               ToBeVisited.push_back(pSuc);
           }

       }

    }
}




void BasicEngine::ClearDijkstra()
{
    for (int y = 0; y < m_nMapHeight; y++)
    {
        for (int x = 0; x < m_nMapWidth; x++)
        {
            if (m_lstCell[x][y]->m_nTraverseCost < 0)
            {
               m_lstCell[x][y]->m_nDijkstra = INT_MAX;
            }
            else
            {
               m_lstCell[x][y]->m_nDijkstra = -1;
            }

        }
    }
}


void BasicEngine::ClearAstar()
{
   m_lstOpenList.erase(m_lstOpenList.begin(),m_lstOpenList.end());
   m_lstClosedList.erase(m_lstClosedList.begin(),m_lstClosedList.end());
   for (int y = 0; y < m_nMapHeight; y++)
    {
        for (int x = 0; x < m_nMapWidth; x++)
        {
            m_lstCell[x][y]->m_pParent  = NULL;
            m_lstCell[x][y]->m_dF       = 0.0;
            m_lstCell[x][y]->m_blIsPath = false;
        }
    }
}


std::vector<BasicTile *> BasicEngine::GetAdjectend(int a_nX, int a_nY, bool blAllowDiagonal)
{
    std::vector<BasicTile *> Result;
    for (int y = a_nY -1; y <= a_nY +1; y++)
    {
        for (int x = a_nX -1; x <= a_nX +1; x++)
        {
             if ((x >= 0) && (x < m_nMapWidth) && (y >= 0) && (y < m_nMapHeight) && ((x != a_nX) || (y != a_nY )))
             {
                 if ((x == a_nX) || (y == a_nY))
                 {
                     if ((m_lstCell[x][y]->m_nTraverseCost > 0) && (!m_lstCell[x][y]->m_blOccupied))
                     {
                        Result.push_back(m_lstCell[x][y]);
                     }
                 }
                 else if (blAllowDiagonal)
                 {
                     if ((m_lstCell[x][y]->m_nTraverseCost > 0) && (!m_lstCell[x][y]->m_blOccupied))
                     {
                        if ((m_lstCell[x][a_nY]->m_nTraverseCost > 0) && (m_lstCell[a_nX][y]->m_nTraverseCost > 0))
                        {
                            Result.push_back(m_lstCell[x][y]);
                        }
                     }
                 }
             }
        }
    }
    return Result;
}







bool BasicEngine::IsInOpen(BasicTile * a_pTile)
{
    bool blFound = false;
    std::list<BasicTile *>::const_iterator p = m_lstOpenList.begin();

    while ((!blFound) && (p != m_lstOpenList.end()))
    {
        if ((*p) == a_pTile)
        {
            blFound = true;
        }
        p++;
    }

    return blFound;
}

bool BasicEngine::IsInClosed(BasicTile * a_pTile)
{
    bool blFound = false;
    std::list<BasicTile *>::const_iterator p = m_lstClosedList.begin();

    while ((!blFound) && (p != m_lstClosedList.end()))
    {
        if ((*p) == a_pTile)
        {
            blFound = true;
        }
        p++;
    }

    return blFound;
}


bool BasicEngine::IsDiagonal(BasicTile * a_pTile1, BasicTile * a_pTile2)
{
    return ((a_pTile1->m_nX != a_pTile2->m_nX) && (a_pTile1->m_nY != a_pTile2->m_nY));
}


bool BasicEngine::LineOfSight(int x1, int y1, int x2, int y2)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;
  bool blBlocked = false;

  if (dxabs >= dyabs) /* the line is more horizontal than vertical */
  {
    i = 0;
    while ((i<dxabs) && (!blBlocked))
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
        if ((m_lstCell[px][py]->m_nTraverseCost <= 0) || (m_lstCell[px][py]->m_blOccupied) ||
            (m_lstCell[px+sdx][py-sdy]->m_nTraverseCost <= 0) || (m_lstCell[px+sdx][py-sdy]->m_blOccupied))
        {
          blBlocked = true;
        }
      }
      px+=sdx;
      if ((m_lstCell[px][py]->m_nTraverseCost <= 0) || (m_lstCell[px][py]->m_blOccupied))
      {
          blBlocked = true;
      }
      i++;
    }
  }
  else /* the line is more vertical than horizontal */
  {
    i = 0;
    while ((i < dyabs) && (!blBlocked))
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
        if ((m_lstCell[px][py]->m_nTraverseCost <= 0) || (m_lstCell[px][py]->m_blOccupied) ||
            (m_lstCell[px-sdx][py+sdy]->m_nTraverseCost <= 0) || (m_lstCell[px-sdx][py+sdy]->m_blOccupied))
        {
          blBlocked = true;
        }
      }
      py+=sdy;
      if ((m_lstCell[px][py]->m_nTraverseCost <= 0) || (m_lstCell[px][py]->m_blOccupied))
      {
          blBlocked = true;
      }
      i++;
    }
  }

  return !blBlocked;
}



bool FSort(const BasicTile * a_pTile1, const BasicTile * a_pTile2)
{
     return (a_pTile1->m_dF < a_pTile2->m_dF);
}





void BasicEngine::DoTest()
{
    for (int y = 0; y < m_nMapHeight; y++)
    {
        for (int x = 0; x < m_nMapWidth; x++)
        {
            std::cout << "Successors of Tile (" << x << "," << y << ")\n";
            std::vector<BasicTile *> vSuccessors;
            vSuccessors = GetAdjectend(m_lstCell[x][y]->m_nX, m_lstCell[x][y]->m_nY);
            for (size_t i=0; i < vSuccessors.size(); i++)                 // loop through each successor
            {
                std::cout << "T (" << vSuccessors[i]->GetX() << "," << vSuccessors[i]->GetY() << ")\n";
            }
            std::cout << "\n";
        }
    }

}




